// jpeg_demo.c
// Simple JPEG capture demo.
// demontrates how to capture JPEGs with the 817 SDK
// with a minimal amount of code.
// Copyright Sensoray Company Inc. 2008
// Author: D.A.

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include "s817.h"
#include "s817f.h"

static int savefile (void *pBuf, int flen, char *fname);
static void get_fw_ver(int board);
static void FrameGrabberQuit(int board, MODE *pMode);

#define MAX_PATH 260

int main(int argc, char **argv )
{
    ECODE ecode;
    char fname[MAX_PATH];
    int aChan;
    unsigned int board = 1;
    MODE mode;  
    MODE defMode = DEFAULT_MODE;
    CAP_STAT *capStat; // pointer to capture status struct
    CAP_BUFFER cBuf;
    int rc;
    int bPAL;
    int captured = 0;
    int bcap[SYS_CHAN_PER_BOARD];  // whether image capture on channel or not
    int numChannels = SYS_CHAN_PER_BOARD;

    memset(bcap, 0, sizeof(bcap));
    if (argc < 2) {
        printf("you must specify the video system\n");
        printf("s817demo system board_num\n");
        printf(" where system 0 ==NTSC, 1 ==PAL\n");
        return -1;
    }
    bPAL = atoi(argv[1]);

    if (argc >= 3) {
        board = atoi(argv[2]);
        printf("using board %u\n", board);
        if ((board < 1) || (board > SYS_BOARDS)) {
            printf("invalid board number!\n");
            printf("enter a board between 1 and %d\n", (int) SYS_BOARDS);
            return -1;
        }
    }

    // our mode settings
    mode = defMode;
    if (bPAL) {
        mode.vMode.tvFormat = TV_PAL;
    }
    mode.cMode.frameRateBmp = FR_FULL;
    mode.cMode.frameRateJpg = FR_FULL;

    // Open all channels
    for (aChan = 1; aChan <= SYS_CHAN_PER_BOARD; aChan++) {
        printf("opening channel %d\n", aChan);
        ecode = S817_OpenChannel(board, aChan, &mode);
        if (ecode != 0) {
            if (ecode == ERR_ILLEGAL_CHAN) {
                numChannels--;
                continue;
            }
            printf ("Chan open error %d; aChan %d\n", ecode, aChan);
            goto eXitus;
        } else {
            ecode = S817_SetVideoMode(board, aChan, &mode.vMode);
            if (ecode != 0) {
                printf ("SetMode error %d chan %d\n", ecode, aChan);
                goto eXitus;
            }
        }
    }

    // set the date and time for the board
    //  (optional): not required if not using OSD
    S817_SetDateTime(board);

    // Start capture on all chan's
    for (aChan = 1; aChan <= SYS_CHAN_PER_BOARD; aChan++) {
        mode.cMode.capMode = CAP_MODE_CONT;
        mode.cMode.capType = CAP_JPG;
        // board, channel, &cMode
        printf("starting channel %d\n", aChan);
        ecode = S817_SetCaptureMode(board, aChan, &mode.cMode);
        if (ecode != 0) {
            if (ecode == ERR_ILLEGAL_CHAN) {
                continue;
            }
            printf ("Capture error %d; channel %d\n", ecode, aChan);
            goto eXitus;
        }
    }

    // get firmware version (optional.  not required for capture)
    get_fw_ver(board);

    captured = 0;
    // grab an image from each channel
    while (captured < numChannels) {
        int capType;
        rc = S817_BlockDone(board, 300, &aChan, &capType);
        if (rc != 0) {
            printf("capture timed out!\n");
            FrameGrabberQuit(board, &mode);
            return 0;
        }

        if (capType != CAP_JPG) {
            // bmp capture should not be turned on.
            // should not get here.
            continue;
        }
        ecode = S817_GetBuffer(board, aChan, &cBuf, CAP_JPG);
        if (ecode != 0) {
            printf ("GetBuffer error %d; aChan %d\n", ecode, aChan);
            return -1;
        }
        capStat = (CAP_STAT *) cBuf.pStat;
        if (capStat->jpgsize && capStat->valid && !bcap[aChan-1]) {
            captured++;
            bcap[aChan-1] = 1;
            sprintf (fname, "snap%02d_%08X", aChan, capStat->tick);
            savefile (cBuf.pData, capStat->jpgsize, fname);
        }
        S817_ReleaseBuffer(board, aChan, cBuf.bufId, CAP_JPG);
    }
 eXitus:
    FrameGrabberQuit(board, &mode);
    printf("done capture\n");
    return 0;
}

// close down the frame grabber cleanly
static void FrameGrabberQuit(int board, MODE *pMode) 
{
    int i, ecode;
    for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
        pMode->cMode.capType = CAP_OFF;
        ecode = S817_SetCaptureMode(board, i + 1, &pMode->cMode);
        if (ecode != 0 && ecode != ERR_ILLEGAL_CHAN) {
            printf ("Error %d calling S817_SetCaptureMode\n", ecode);
        }
    }
    return;
}

// save JPEG to file
static int savefile(void *pBuf, int flen, char *fname)
{
    FILE *fin;
    int ecode = 0;
    char str[MAX_PATH];
    int rc;
    sprintf (str, "%s.jpg", fname);
    printf("savefile %s, len:%d\n",(char *)str, flen);
    fin = fopen( str, "wb+");
    if (fin != NULL) {
        rc = fwrite( pBuf, 1, flen, fin);
        if (rc <= 0) {
            printf("write file error\n");
            ecode = -1;
        }
        fclose(fin);
    } else {
        printf("error\n");
        ecode = -1;
    }
    return ecode;
}

// for debug.  retrieves the fw version in the system
static void get_fw_ver(int board)
{
    SYS_STATUS sysStat;
    ECODE ecode;
    ecode = S817_GetStatusInfo(board, 0, &sysStat);
    if (ecode != 0) {
        printf("Error %d calling S817_GetStatusInfo\n", ecode);
        return;
    }
    printf("s817 fw version %u.%u\n", (sysStat.fwVer >> 16 & 0xff), (sysStat.fwVer & 0xff));
}
