// s817cmd.c
// The purpose of this demo is to exercise the 817 board.
// jpeg_demo.c is a simpler demo that shows how to capture with less code.
//
// Copyright (C) 2007-2009  Sensoray Co., Inc.  This program is free software;
// Linux Command line demo for 817.
// No X windows required. No openGL required
// For evaluation of the 817 without display.
// To determine FPS and 817 capabilities without the complications of
// displaying images.
// Demonstrates multithreaded usage of the 817 API/SDK
//
// you can redistribute it and/or modify it under the terms of the GNU
// General Public License as published by the Free Software Foundation;
// either version 2 of the License, or (at your option) any later version.
// This program is distributed in the hope that it will be useful,
// but WITHOUT ANY WARRANTY; without even the implied warranty of
// MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
// GNU General Public License for more details.
//
// You should have received a copy of the GNU General Public License
// along with this program; if not, write to the Free Software
// Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
// 
// You must obey the GNU General Public License in all respects for all of
// the code used. If you modify this file, you may extend this exception
// to your version of the file, but you are not obligated to do so. If you
// do not wish to do so, delete this exception statement from your version.

#include <stdlib.h>
#include <stdio.h>
#include <string.h>
#include <sys/time.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <unistd.h>
#include <fcntl.h>
#include <limits.h>
#include <pthread.h>
#include <sys/time.h>
#include <time.h>
#include <assert.h>
#include <ctype.h>
#include <errno.h>
#include "s817.h"
#include "s817f.h"

#define MAXLEN              80 // Maximum length of user command string.
#define MODE_IDLE           0  // stream mode
#define MODE_STREAM         1  // stream mode
#define MAX_CHANNELS        SYS_CHAN_PER_BOARD
#define FIFO_NAMEBASE_JPG "s817_fifo_jpg"
#define FIFO_NAMEBASE_BMP "s817_fifo_bmp"

typedef struct BOARD {                          // BOARD OBJECT -------
    int             smode;        // Stream mode.
    int             stop_rq;      // Stream Halt requested.
    char            fname[MAXLEN];              // File name for snapshots
    pthread_t       io_thread_idM; // Handle to stream i/o thread.
    pthread_t       io_thread_idJ; // Handle to jpeg processing thread.
    pthread_t       io_thread_idB; // Handle to bmp processing thread.
    unsigned int    frame_stats_jpg[MAX_CHANNELS];  // Number of frames(used for FPS calc only)
    unsigned int    frame_stats_bmp[MAX_CHANNELS];  // 
    unsigned int    jpg_frame_count[MAX_CHANNELS];
    unsigned int    bmp_frame_count[MAX_CHANNELS];
    int             snapshot_ppm[MAX_CHANNELS];
    int             snapshot_jpeg[MAX_CHANNELS];
    MODE            mode[MAX_CHANNELS];         // 817 Mode structure
    int             board;                      // board number
    int             print_stats;
    int             b_stat_exclude[MAX_CHANNELS]; // don't update stats if just started
    int             jpeg_load[MAX_CHANNELS];
    int             jpegs_saved[MAX_CHANNELS];
    char            fifo_bmp[260]; //fifo file name
    char            fifo_jpg[260];
} BOARD;

struct io_thread_param {
    BOARD   *b;
};

// forward declarations
static int app_init(int );
static void app_shutdown( void );
static void cmd_processor( void );
static void *main_thread_func( void* arg );
static void *jpg_thread_func( void* arg );
static void *bmp_thread_func( void* arg );
static void get_attributes( struct BOARD *b );
static void start_stream( struct BOARD *b, int chan, int capType );
static void stop_stream( struct BOARD *b, int chan );
static void stop_threads(struct BOARD *b);
static int start_threads(struct BOARD *b);
static void print_ecode(ECODE ecode);
static void getPrintErrorInfo(struct BOARD *b);
static void prompt( void );
static void print_stats(BOARD *b, double seconds);
static double current_time(void);
static int save_image_ppm(const unsigned char *image, const char *szFilename, int width, int height);
static int save_jpg_file(void *pBuf, int flen, char *fname, int bloud);
static void getSysStatus(struct BOARD *b);

// globals
static int G_fwver = 0;
static int G_verbose = 0;
static int G_rsrv[MAX_CHANNELS];  // debug only(not required)
static struct BOARD bd;     // Board objects.
static int act_bdnum   = 0;    // Board 0 is active by default.

// BMP grabbing thread
static void *bmp_thread_func( void *arg )
{
    BOARD                  *b;
    unsigned char          *rgb_frame; // for snapshots
    struct io_thread_param *param = (struct io_thread_param *) arg;
    CAP_BUFFER      cBuf;
    int rc;
    int fifo_fd;
    CAP_STAT *capStat;
    int aChan;
    int board;

    b = param->b;
    board = b->board;
    free( arg);
    rgb_frame = (unsigned char *) malloc(768*576*3);

    if (rgb_frame == NULL) {
        printf("out of memory\n");
        return NULL;
    }

    fifo_fd = open(b->fifo_bmp, O_RDONLY);
    if (fifo_fd == -1) {
        printf("could not open fifo\n");
        free(rgb_frame);
        return NULL;
    }

    while (!b->stop_rq) {
        rc = read(fifo_fd, &cBuf, sizeof(CAP_BUFFER));
        if (rc <= 0)
            continue;

        capStat = (CAP_STAT *) cBuf.pStat;
        aChan = capStat->chanId;

        if (aChan < 1 || aChan > MAX_CHANNELS)
            continue;

        if (capStat->chanId != aChan)
            printf ("Chan ID error on chan %d: %d\n", aChan, capStat->chanId);

        // if BMP snapshot triggered, save to BMP
        if (b->snapshot_ppm[aChan-1]) {
            int bPAL = b->mode[aChan-1].vMode.tvFormat == TV_PAL ? 1 : 0;
            int width = bPAL ? LINE_PAL_FULL : LINE_NTSC_FULL;
            int height = bPAL ? LINES_PER_FIELD_PAL * 2
                : LINES_PER_FIELD_NTSC * 2;
            switch (b->mode[aChan-1].vMode.jpgSize) {
            case SIZE_4CIF:
            case SIZE_4CIFI:
            default:
                break;
            case SIZE_2CIF:
                height /= 2;
                break;
            case SIZE_1CIF:
                height /= 2;
                width /=2;
                break;
            }
            save_image_ppm((const unsigned char *) cBuf.pData,
                           b->fname, width, height);
            prompt();
            b->snapshot_ppm[aChan-1] = 0;
        }
        G_rsrv[aChan-1] = capStat->reserved[1]; //debug only
        // release the buffer
        rc = S817_ReleaseBuffer(board, aChan, cBuf.bufId, CAP_BMP);
        if (rc != 0) {
            printf("failed to release buffer\n");
        }
    }
    close(fifo_fd);
    free(rgb_frame);
    return NULL;
}

// jpg grabbing thread
static void *jpg_thread_func( void *arg )
{
    BOARD *b;
    struct io_thread_param *param = (struct io_thread_param *) arg;
    CAP_STAT *capStat;
    int rc;
    int board;
    int fifo_fd;
    CAP_BUFFER cBuf;
    int aChan;
    b = param->b;
    board = b->board;
    free(arg);
    fifo_fd = open(b->fifo_jpg, O_RDONLY);
    if (fifo_fd == -1) {
        printf("could not open fifo jpg\n");
        return NULL;
    }

    while (!b->stop_rq) {
        rc = read(fifo_fd, &cBuf, sizeof(CAP_BUFFER));
        if (rc <= 0) {
            continue;
        }
        capStat = (CAP_STAT *) cBuf.pStat;
        aChan = capStat->chanId;
        // if JPEG snapshot triggered save the jpg
        if (b->snapshot_jpeg[aChan-1]) {
            if (capStat->jpgsize && capStat->valid &&
                (b->mode[aChan-1].cMode.capType & CAP_JPG)) {
                save_jpg_file (cBuf.pData, capStat->jpgsize, b->fname, 1);
            }
            b->snapshot_jpeg[aChan-1] = 0;
        }

        if (b->jpeg_load[aChan-1]) {
            char fname[100];
            sprintf(fname, "%01d-%02d.jpg", b->board, aChan);
            if (capStat->jpgsize && capStat->valid &&
                (b->mode[aChan-1].cMode.capType & CAP_JPG)) {
                save_jpg_file (cBuf.pData, capStat->jpgsize, fname, 0);
                b->jpegs_saved[aChan-1]++;
            }
        }
        G_rsrv[aChan-1] = capStat->reserved[1]; //debug only
        // we are done with the buffer.  We must release it
        // now so the driver can reuse it
        rc = S817_ReleaseBuffer(board, aChan, cBuf.bufId, CAP_JPG);
        if (rc != 0) {
            printf("release buffer JPG failed RC:%d"
                   "chan:%d bufID:%d\n", rc, aChan, cBuf.bufId);
        }
    }
    close(fifo_fd);
    return NULL;
}

// Capture thread. Where frames are captured.
static void *main_thread_func( void *arg )
{
    BOARD                  *b;
    struct io_thread_param *param = (struct io_thread_param *) arg;
    CAP_BUFFER      cBuf;
    ECODE ecode;
    int rc;
    int board;
    int fifo_fd1, fifo_fd2;
    int aChan, capType;
    double t = current_time();
    double tRot0;
    double tRate0 = -1.0;

    b = param->b;
    board = b->board;
    free( arg);

    fifo_fd1 = open(b->fifo_bmp, O_WRONLY);
    if (fifo_fd1 == -1) {
        printf("could not open BMP fifo\n");
        return NULL;
    }
    fifo_fd2 = open(b->fifo_jpg, O_WRONLY);
    if (fifo_fd2 == -1) {
        printf("could not open JPG fifo\n");
        return NULL;
    }

    while (!b->stop_rq) {
        rc = S817_BlockDone(board, 200, &aChan, &capType);
        if (rc != 0) {
            continue;
        }
        // grab the buffer(s)
        if (capType & CAP_BMP) {
            b->bmp_frame_count[aChan - 1]++;
            ecode = S817_GetBuffer(board, aChan, &cBuf, CAP_BMP);
            if (ecode) {
                printf("failed to get buffer\n");
            }
            rc = write(fifo_fd1, &cBuf, sizeof(CAP_BUFFER));
            if (rc <= 0) {
                printf("failed to save fifo\n");
            }
            b->frame_stats_bmp[aChan - 1]++;

        }
        if (capType & CAP_JPG) {
            b->jpg_frame_count[aChan - 1]++;
            ecode = S817_GetBuffer(board, aChan, &cBuf, CAP_JPG);
            if (ecode) {
                printf("failed to get buffer\n");
            }
            rc = write(fifo_fd2, &cBuf, sizeof(CAP_BUFFER));
            if (rc <= 0) {
                printf("failed to save fifo\n");
            }
            b->frame_stats_jpg[aChan - 1]++;
        }
        t = current_time();
        // begin timing information (optional)
        if (tRot0 < 0.0)
            tRot0 = t;
        tRot0 = t;
        if (tRate0 < 0.0)
            tRate0 = t;
        if (t - tRate0 >= 3.0) {
            int c;
            double seconds = t - tRate0;
            if (b->print_stats) {
                print_stats(b, seconds);
                tRate0 = t;
                // reset all the frames
                for (c = 0; c < SYS_CHAN_PER_BOARD; c++) {
                    b->frame_stats_bmp[c] = 0;
                    b->frame_stats_jpg[c] = 0;

                }
            }
        }
        // end statistics information
    }
    // finish up 
    b->smode = MODE_IDLE;
    close(fifo_fd1);
    close(fifo_fd2);
    printf( "Stream stopped.\n");
    prompt();
    return NULL;
}


// Init the system and execute user commands until termination.
int main( int argc, char *argv[] )
{
    int rtnval;

    if (argc < 2) {
        printf("usage: ./s817demo board_number(1..8)\n");
        return 0;
    }
    // Attempt to open and initialize all boards.
    rtnval = app_init(atoi(argv[1]));

    // If all boards successfully opened and initialized ...
    if (rtnval == 0 ) {
        // Execute user commands until Application Quit command is invoked.
        cmd_processor();
    }
    // Close all opened boards.
    app_shutdown();
    // Exit application.
    return 0;
}

// Open and initialize one board.  Returns: 0=success, -1=error.
static int board_open(int board_num)
{
    struct BOARD *b = &bd;
    MODE defmode = DEFAULT_MODE;
    int aChan;
    ECODE ecode;
    SYS_STATUS st;
    int numChannels = SYS_CHAN_PER_BOARD;
    // open all channels on the board. This is done
    // at the start of the application for all channels
    // we want to use later.
    b->smode = MODE_IDLE;
    b->fname[0] = 0;
    b->board = board_num;
    for (aChan = 1; aChan <= SYS_CHAN_PER_BOARD; aChan++) {
        printf("opening channel %d\n", aChan);
        b->mode[aChan - 1] = defmode;
        ecode = S817_OpenChannel(board_num, aChan, &b->mode[aChan - 1]);
        if (ecode != 0) {
            if (aChan > 1 && ecode == ERR_ILLEGAL_CHAN) {
                numChannels--;
                printf ("Chan open error %d; aChan %d is not present\n", ecode, aChan);
                ecode = 0;
                continue;
            }
            if ((board_num != 1 && aChan != 1) || (board_num == 1)) {
                printf ("Chan open error %d; aChan %d\n", ecode, aChan);
                getPrintErrorInfo(b);
            }
            return ecode;
        } else {
            ecode = S817_SetVideoMode(board_num, aChan, &b->mode[aChan - 1].vMode);
            if (ecode != 0) {
                printf ("SetMode error %d chan %d\n", ecode, aChan);
                getPrintErrorInfo(b);
                return ecode;
            }
        }
    }
    printf("done opening %d channels.\n", numChannels);
    S817_SetDateTime(board_num);
    if (start_threads(b) != 0)
        return -EFAULT;
    S817_GetStatusInfo(board_num, 0, &st);
    G_fwver = st.fwVer;
    return ecode;
}


///////////////////////////////////////////////////////////////////////
// Open and initialize all boards.  Returns: -1=error, 0=success.

static int app_init(int board_num )
{
    printf("opening board number %d\n", board_num);
    // Open device.  Return error if open failed.
    if (board_open(board_num ) != 0 ) {
        return -1;
    }
    act_bdnum = board_num - 1;
    printf("opened board %d\n", board_num);
    return 0;
}

//////////////////////////////////////////////////////////////
// Terminate all stream threads and close all boards.

static void app_shutdown(void)
{
    struct BOARD *b;
    int i;
    int j;
    ECODE ecode;
    // For each board ...
    i = act_bdnum;
    b = &bd;
    stop_threads(b);
    // Terminate the board's I/O thread if it exists.
    for(j = 1; j <= SYS_CHAN_PER_BOARD; j++) {
        stop_stream( b, j);
    }
    // Close the board.
    printf("closing board %d\n", i + 1);
    for (j = 1; j <= SYS_CHAN_PER_BOARD; j++) {
        ecode = S817_CloseChannel(i + 1, j);
        if (ecode != 0 && ecode != ERR_ILLEGAL_CHAN && ecode != ERR_CHAN_NOTOPEN) {
            printf("failed to close channel %d, ecode %d\n", j, ecode);
        }
    }
    printf( "\n" );
}

// display the prompt
static void prompt( void )
{
    printf( "s817demo[%d]>", act_bdnum + 1);
    fflush( stdout );
}

// Display the command menu.
static void show_menu( void )
{
    printf("+-S817DEMO-COMMANDS------------------------------------------------------------+\n");
    printf("|   p n ........... print stats (1=on, 0 off)                                  |\n");
    printf("|   get ........... display current settings                                   |\n");
    printf("|   h[elp] ........ display this list                                          |\n");
    printf("|   q[uit] ........ quit program                                               |\n");
    printf("+-VIDEO------------------------------------------------------------------------+\n");
    printf("|   bright c v .... set luminance brightness v: 0..255 for channel c           |\n");
    printf("|   contrast c v ...set luminance contrast v: 0..255 for channel c             |\n");
    printf("|   vsat c v ...... set chrominance saturation v: 0..255 for channel c         |\n");
    printf("|   vhue c v ...... set hue v: 0..255 for channel c                            |\n");  
    printf("|   vsystem c v.... set video system v: 0=NTSC, 1=PAL for channel c            |\n");
    printf("|   bdec c v....... set frame rate for BMP c=channel, v=value                  |\n");
    printf("|   jdec c v....... set frame rate for JPG c=channel, v=value                  |\n");
    printf("                    (0 all frames, 1=4/5, 2=2/3, 3=1/2, 4=2/5, 5=1/3, 6=1/5    |\n");
    printf("                    (7=1/10, 8=1/15)                                           |\n");
    printf("|   scale c v...... set scale c=channel, v=value(0-4CIFS, 1-2CIFS, 2-1CIFS)    |\n");
    printf("|                   3-4CIFI(interpolated)                                      |\n");
    printf("|   jqual c v...... set jpeg quality c=channel, v=value(0-100)                 |\n");
    printf("|   osd c v........ toggle osd timestamp (v=1 on)                              |\n");
    printf("+-STREAM-----------------------------------------------------------------------+\n");
    printf("|   start c m...... start stream c = channel(1-16) (-1 for all)                |\n");
    printf("|                   m = optional mask (1=JPG(default), 2 = BMP, 3 = BMP+JPG)   |\n");
    printf("|   stop c m....... stop stream c = channel(1-16) (-1 for all)                 |\n");
    printf("|                   m = optional mask (1=BMP(default), 2 = JPG, 3 = BMP+JPG)   |\n");
    printf("|   ppm c f........ takes raw snapshot on c channel(1-16) to file f.ppm        |\n");
    printf("|                   BMP must be turned on(see mask in start command)           |\n");
    printf("|   jpeg c f....... takes jpeg snapshot on channel c = channel(1-16) to file f |\n");
    printf("|                   jpeg must be turned on(see mask in start command)          |\n");
    printf("|   jload c ....... jpeg test mode on channel c(-1 for all).                   |\n");
    printf("|                   takes continuous jpeg snapshots overwriting previous file. |\n");
    printf("|                   file is b-cc.jpg where b is the board, cc is the channel.  |\n");
    printf("|                   Exit jpeg test by stopping the stream.                     |\n");
    printf("|   err ........... print extended debug information                           |\n");
    printf("|   verbose........ verbose debug                                              |\n");
    printf("+------------------------------------------------------------------------------+\n");
    printf("fw version: %d.%d\n", G_fwver >> 16, G_fwver & 0xff);
}



// Decode and execute user command.  Returns TRUE if Quit selected.
static int exec_command( char *cmd, char *arg,char *arg2 )
{
    struct BOARD    *b;
    // Convert argument to integer value (not relevant for all commands).
    int arg_val = -1;
    sscanf( arg, "%d", &arg_val );
    // Terminate application.
    if (cmd[0] == 'q' )
        return 1;
    // Show help.
    if ((cmd[0] == 'h') || (cmd[0] == '?')) {
        show_menu();
        prompt();
        return 0;
    }
    // Prompt for user command.
    prompt();
    // Set pointer to active board's object.
    b = &bd;
    if (strcmp(cmd, "get") == 0 ) {
        get_attributes(b);
    } else if (strcmp( cmd, "start"       ) == 0 ) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        if ((chn < 1 || chn > SYS_CHAN_PER_BOARD) && (chn !=-1)) {
            printf("invalid channel\n");
            prompt();
            return 0;
        }
        if (arg2[0] == 0) {
            val = CAP_JPG;
        }
        start_stream( b, chn, val);
        printf("type \"p\" followed by \"Enter\" for stats\n");
        prompt();
    } else if (strcmp( cmd, "stop"      ) == 0 ) {
        int chn,val;
        int j;
        sscanf( arg, "%d", &chn);
        sscanf( arg2, "%d", &val);
        if ((chn < 1 || chn > SYS_CHAN_PER_BOARD) && (chn !=-1)) {
            printf("invalid channel\n");
            prompt();
            return 0;
        }
        stop_stream( b, (arg[0] == 0) ? 1 : arg_val);
        for (j = 0; j < SYS_CHAN_PER_BOARD; j++) {
            if (b->jpeg_load[j]) {
                printf("saved %d JPEGs channel %02d\n", b->jpegs_saved[j], j+1);
            }
        }
        memset(b->jpeg_load, 0, sizeof(int) * MAX_CHANNELS);
        memset(b->jpegs_saved, 0, sizeof(int) * MAX_CHANNELS);
        prompt();
    } else if (strcmp( cmd, "ppm") == 0 ) {
        int chn;
        sscanf( arg, "%d",&chn);
        if (chn < 1 || chn > SYS_CHAN_PER_BOARD)
            chn = 1;
        strcpy(b->fname, ( arg2[0] == 0 ) ? "snapshot.ppm" : arg2);
        if ((b->mode[chn-1].cMode.capType & CAP_BMP) == 0) {
            printf("stream must be running with BMP capture!\n");
            prompt();
        }
        else {
            // fire snapshot in IO thread
            b->snapshot_ppm[chn-1] = 1;
        }
    } else if ((strcmp( cmd, "jpg") == 0) || (strcmp( cmd, "jpeg") == 0)) {
        int chn;
        sscanf( arg, "%d",&chn);
        if (chn < 1 || chn > SYS_CHAN_PER_BOARD)
            chn = 1;
        strcpy(b->fname, ( arg2[0] == 0 ) ? "snapshot.jpg" : arg2);
        if ((b->mode[chn-1].cMode.capType & CAP_JPG) == 0) {
            printf("stream must be running with JPG capture!\n");
            prompt();
        }
        else {
            // fire snapshot in IO thread
            b->snapshot_jpeg[chn-1] = 1;
        }
    } else if (strcmp( cmd, "p") == 0 ) {
        int val;
        sscanf( arg, "%d",&val);
        b->print_stats = val;
    } else if (strcmp( cmd, "jload") == 0 ) {
        int chn;
        int j;
        sscanf( arg, "%d",&chn);
        if ((chn < 1 || chn > SYS_CHAN_PER_BOARD) && (chn !=-1)) {
            printf("invalid channel\n");
            prompt();
            return 0;
        }
        if (chn == -1) {
            for (j=0;j<SYS_CHAN_PER_BOARD;j++) {
                b->jpeg_load[j] = 1;
                if (b->mode[j].cMode.capType == CAP_OFF) {
                    printf("Use \"start\" command to begin test\n");
                }
            }
        } else {
            b->jpeg_load[chn] = 1;
            if (b->mode[chn].cMode.capType == CAP_OFF) {
                printf("Use \"start\" command to begin test\n");
            }
        }
    } else if (strcmp( cmd, "bdec" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (((chn > SYS_CHAN_PER_BOARD || chn < 1) && (chn != -1)) || val < 0 || val > 8) {
            printf("bad channel or decimation value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            if (chn == -1 ) {
                for (chn=1;chn<=SYS_CHAN_PER_BOARD;chn++) {
                    b->mode[chn-1].cMode.frameRateBmp = val;
                    ret = S817_SetCaptureMode( b->board, chn, &b->mode[chn-1].cMode);
                    if (ret != 0 ) {
                        printf("failed set mode 0x%x\n", ret);
                        prompt();
                    }
                }
            } else {
                b->mode[chn-1].cMode.frameRateBmp = val;
                ret = S817_SetCaptureMode( b->board, chn, &b->mode[chn-1].cMode);
                if (ret != 0 ) {
                    printf("failed set mode 0x%x\n", ret);
                    prompt();
                }
            }
        }
    } else if (strcmp( cmd, "jdec" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (((chn > SYS_CHAN_PER_BOARD || chn < 1) && (chn != -1)) || val < 0 || val > 8) {
            printf("bad channel or decimation value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            if (chn == -1 ) {
                for (chn=1;chn<=SYS_CHAN_PER_BOARD;chn++) {
                    b->mode[chn-1].cMode.frameRateJpg = val;
                    ret = S817_SetCaptureMode( b->board, chn, &b->mode[chn-1].cMode);
                    if (ret != 0 ) {
                        printf("failed set mode 0x%x\n", ret);
                        prompt();
                    }
                }
            } else {
                b->mode[chn-1].cMode.frameRateJpg = val;
                ret = S817_SetCaptureMode( b->board, chn, &b->mode[chn-1].cMode);
                if (ret != 0 ) {
                    printf("failed set mode 0x%x\n", ret);
                    prompt();
                }
            }
        }
    } else if (strcmp( cmd, "scale" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > SYS_CHAN_PER_BOARD || chn < 1 || val <0 || val > 3) {
            printf("bad channel or scale value %d %d\n",chn,val);
            prompt();
        } else {
            int ret;
            b->mode[chn-1].vMode.jpgSize = val;
            ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "bright" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > SYS_CHAN_PER_BOARD || chn < 1 || val <0 || val > 255) {
            printf("bad channel or bright value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].vMode.brightness = val;
            ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "contrast" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > SYS_CHAN_PER_BOARD || chn < 1 || val <0 || val > 255) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].vMode.contrast = val;
            ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "vsat" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > SYS_CHAN_PER_BOARD || chn < 1 || val <0 || val > 255) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        } else {
            int ret;
            b->mode[chn-1].vMode.saturation = val;
            ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "vhue" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > SYS_CHAN_PER_BOARD || chn < 1 || val <0 || val > 255) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].vMode.hue = val;
            ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "jqual" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > SYS_CHAN_PER_BOARD || chn < 1 || val <0 || val > 100) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        }
        else {
            int ret;
            b->mode[chn-1].cMode.compQuality = val;
            ret = S817_SetCaptureMode( b->board, chn, &b->mode[chn-1].cMode);
            if (ret != 0 ) {
                printf("failed set mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "osd" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > SYS_CHAN_PER_BOARD || chn < 1 || val <0 || val > 1) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        } else {
            OSD_MODE osd;
            int ret;
            memset(&osd, 0, sizeof(osd));
            osd.osdOn = val;
            osd.osdBmp = 1;
            osd.transparent = 1;
            osd.positionTop = 1;
            osd.ddmm = 1;
            osd.year2 = 0;
            osd.fraction = 2;
            strncpy(osd.line, "Date: ^d Time: ^t", 80);
            ret = S817_SetOsdMode(b->board, chn, &osd);
            if (ret != 0 ) {
                printf("failed set OSD mode\n");
                prompt();
            }
        }
    } else if (strcmp( cmd, "vsystem" ) == 0) {
        int chn,val;
        sscanf( arg, "%d",&chn);
        sscanf( arg2, "%d",&val);
        // check the channel
        if (chn > SYS_CHAN_PER_BOARD || chn < 1 || val < 0 || val >= TV_MAX) {
            printf("bad channel or value %d %d\n",chn,val);
            prompt();
        } else {
            int ret;
            b->mode[chn-1].vMode.tvFormat = val;
            if (b->mode[chn-1].cMode.capType != CAP_OFF) {
                printf("You must stop the stream first before changing the video system\n");
                prompt();
            } else {
                ret = S817_SetVideoMode( b->board, chn, &b->mode[chn-1].vMode);
                if (ret != 0 ) {
                    printf("failed set video system\n");
                    prompt();
                } else {
                    printf("changed to %s on channel %d\n", val == 0 ? "NTSC" : "PAL",
                           chn);
                }
            }
        }
    } else if (strcmp(cmd, "err") == 0) {
        getPrintErrorInfo(b);
        prompt();
    } else if (strcmp(cmd, "verbose") == 0) {
        printf("verbose debugging on\n");
        G_verbose = 1;
        prompt();
    } else if (cmd[0] != 0 ) {
        printf( "Invalid command: %s\n", cmd );
        prompt();
    }
    return 0;
}

// Execute user commands until Application Quit is requested.
static void cmd_processor( void )
{
    char                    cmdbuf[MAXLEN];
    char                    cmd[MAXLEN];
    char                    arg[MAXLEN];
    char                    arg2[MAXLEN];
    char                    *c;
    // Display the command menu.
    show_menu();
    // Prompt for user command.
    prompt();
    // Repeat until Application Terminate is requested ...
    for ( ; ; ) {
        // Input user command string, and if no errors encountered ...
        if (fgets( cmdbuf, MAXLEN - 1, stdin ) != NULL ) {
            // Truncate newline char at end of command string.
            cmdbuf[ strlen( cmdbuf ) - 1 ] = 0;
            // If command line is not blank ...
            if (cmdbuf[0] != 0) {
                // Force command string to lower case, except command=="text".
                if (strncmp( cmdbuf, "text", 4 ) != 0 )
                    for ( c = cmdbuf; *c != 0; c++ ) 
                        *c = tolower( *c );

                // Parse command string into opcode and argument.
                cmd[0] = arg[0] = 0;
                arg2[0] = 0;
                sscanf( cmdbuf, "%s %s %s", cmd, arg,arg2 );
                // Decode and execute command
                if (exec_command( cmd, arg,arg2))
                    break;
            } else {
                // Prompt for user command.
                prompt();
            }
        }
    }
}

// Fetch and display the attributes.
static void get_attributes( struct BOARD *b )
{
    int i;
    char status[100];
    char tmp[100];
    int bPAL;
    int fr_bmp;
    int fr_jpg;
    char *fr_ntsc[9] = { "30fps", "24fps", "20fps", "15fps", "12fps",
                         "10fps", "6fps", "3fps", "2fps"};
    char *fr_pal[9] = { "25fps", "20fps", "16.7fps", "12.5fps", "10fps",
                        "8.3fps", "5fps", "2.5fps", "1.7fps"};

    printf( "\nCURRENT SETTINGS -----------\n" );
    printf( "\nBoard %d----------\n", b->board);
    for (i=0;i<SYS_CHAN_PER_BOARD;i++) {
        bPAL = b->mode[i].vMode.tvFormat == TV_PAL ? 1 : 0;
        fr_bmp = b->mode[i].cMode.frameRateBmp;
        fr_jpg = b->mode[i].cMode.frameRateJpg;
        switch (b->mode[i].cMode.capType) {
        case CAP_BMP:
            sprintf(status, "BMP(%s)", bPAL ? fr_pal[fr_bmp] : fr_ntsc[fr_bmp]);
            break;
        case CAP_JPG:
            sprintf(status, "JPG(%s)", bPAL ? fr_pal[fr_bmp] : fr_ntsc[fr_bmp]);
            break;
        case (CAP_JPG | CAP_BMP):
            sprintf(status, "BMP(%s)+JPG(%s)",  bPAL ? fr_pal[fr_bmp] : fr_ntsc[fr_bmp],
                    bPAL ? fr_pal[fr_jpg] : fr_ntsc[fr_jpg]);
            break;
        case CAP_OFF:
        default:
            sprintf(status, "OFF");
            break;
        }
        snprintf(tmp, 100, "channel %d, capture %s", i, status);
        puts(tmp);
    }
}


static int start_threads(struct BOARD *b)
{
    pthread_attr_t attr;
    struct io_thread_param *param;
    int rc;
    pthread_attr_init( &attr );
    pthread_attr_setdetachstate( &attr, PTHREAD_CREATE_DETACHED );
    param = (struct io_thread_param *) malloc( sizeof(struct io_thread_param));//freed by thread
    if (param == NULL) {
        printf("out of memory \n");
        return -ENOMEM;
    }

    if (b->smode != MODE_IDLE) {
        // Abort command.
        printf( "already running\n" );
        prompt();
        return 0;
    }
    sprintf(b->fifo_bmp, "%s_%d", FIFO_NAMEBASE_BMP, b->board);
    sprintf(b->fifo_jpg, "%s_%d", FIFO_NAMEBASE_JPG, b->board);
    if (access(b->fifo_bmp, F_OK) == -1) {
        rc = mkfifo(b->fifo_bmp, 777);
        if (rc != 0) {
            printf("could not create fifo\n");
            return -ENOMEM;
        }
    }
    if (access(b->fifo_jpg, F_OK) == -1) {
        rc = mkfifo(b->fifo_jpg, 777);
        if (rc != 0) {
            printf("could not create fifo\n");
            return -ENOMEM;
        }
    }

    param->b = b;
    b->smode = MODE_STREAM;
    prompt();
    pthread_create( &b->io_thread_idM, NULL, main_thread_func, param );
    param = (struct io_thread_param *) malloc( sizeof(struct io_thread_param));//freed by thread
    if (param == NULL) {
        printf("out of memory \n");
        return -ENOMEM;
    }
    param->b = b;
    b->smode = MODE_STREAM;
    pthread_create( &b->io_thread_idJ, NULL, jpg_thread_func, param );
    param = (struct io_thread_param *) malloc( sizeof(struct io_thread_param));  //freed by thread
    if (param == NULL) {
        printf("out of memory \n");
        return -ENOMEM;
    }
    param->b = b;
    b->smode = MODE_STREAM;
    pthread_create( &b->io_thread_idB, NULL, bmp_thread_func, param );
    pthread_attr_destroy( &attr );
    return 0;
}

static void stop_threads(struct BOARD *b)
{
    if (b->smode != MODE_IDLE ) {
        // Request thread termination.
        b->stop_rq = 1;
        // Wait for the threads to terminate
        pthread_join( b->io_thread_idM, NULL);
        pthread_join( b->io_thread_idJ, NULL);
        pthread_join( b->io_thread_idB, NULL);
    }
    return;
}


static void start_stream_inner( struct BOARD *b, int chan, int mask)
{
    MODE *pmode = &b->mode[chan-1];
    ECODE ecode;
    char *mode_str[] = {"JPG", "BMP", "JPG+BMP" };
    // start capture for this channel
    pmode->cMode.capMode = CAP_MODE_CONT;
    pmode->cMode.capType = mask;
    // board, channel, &cMode
    ecode = S817_SetCaptureMode(b->board, chan, &pmode->cMode);
    if (ecode) {
        printf("Capture error %d; channel %d board %d\n", ecode, chan, b->board);
        print_ecode(ecode);
        pmode->cMode.capType = CAP_OFF;
        return;
    }
    printf("starting channel %d in %s mode\n", chan, mode_str[pmode->cMode.capType-1]);
    b->b_stat_exclude[chan-1] = 1;
    return;
}

static void start_stream( struct BOARD *b, int chan, int mask )
{
    int i;
    i = chan;
    if (chan <= 0 || chan > SYS_CHAN_PER_BOARD) {
        printf("starting all channels\n");
        prompt();
        for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
            start_stream_inner(b, i+1, mask);
        }

    } else {
        start_stream_inner(b, chan, mask);
    }
    return;
}

// Halt data stream
static void stop_stream_inner( struct BOARD *b, int chan )
{
    MODE *pmode = &b->mode[chan-1];
    ECODE ecode;
    pmode->cMode.capType = CAP_OFF;
    ecode = S817_SetCaptureMode(b->board, chan, &pmode->cMode);
    if (ecode != 0 && ecode != ERR_ILLEGAL_CHAN) {
        printf("Capture error %d; channel %d\n", ecode, chan);
        prompt();
        return;
    }

}

static void stop_stream( struct BOARD *b, int chan )
{
    if (chan == -1) {
        int i;
        printf("stopping all channels\n");
        prompt();
        for (i = 0; i < SYS_CHAN_PER_BOARD; i++)
            stop_stream_inner(b, i+1);
    } else
        stop_stream_inner(b, chan);
    return;
}

// return current time (in seconds)
static double current_time(void)
{
    struct timeval tv;
    struct timezone tz;
    (void) gettimeofday(&tv, &tz);
    return (double) tv.tv_sec + tv.tv_usec / 1000000.0;
}



static void print_stats(BOARD *b, double seconds)
{
    int i;

    if (seconds == 0)
        return;
    printf("\n====================================\n");
    printf("current stats: (Enter \"p 0\" to turn off stats)\n");
    printf("last update %lf seconds (type h for help)\n", seconds);
    for (i = 0; i < SYS_CHAN_PER_BOARD; i++) {
        double fps_jpg = b->frame_stats_jpg[i] / seconds;
        double fps_bmp = b->frame_stats_bmp[i] / seconds;

        if (b->mode[i].cMode.capType == CAP_OFF)
            continue;

        if (b->b_stat_exclude[i]) {

            printf("chan %d: %u JPG frames(just started, no fps calculation)\n", i + 1,
                   b->frame_stats_jpg[i]);
            b->b_stat_exclude[i] = 0;

        } else {
            printf("chan %d: %u(bmp) %u(jpg) frames, %6.3f fps(bmp) %6.3f fps(jpg)", i + 1,
                   b->frame_stats_bmp[i],
                   b->frame_stats_jpg[i],
                   fps_bmp,fps_jpg);
            if (G_verbose)
                printf(" rsrvd[1]: 0x%02x", G_rsrv[i]);
            printf("\n");
        }
    }
    printf("====================================\n");
    prompt();
    return;
}

// error codes to text
static void print_ecode(ECODE ecode)
{
    switch (ecode) {
    case ERR_SX_OPEN:
        printf("unable to open the driver\n");
        break;
    case ERR_BRIDGE_NOT_FOUND:
        printf("no 817 bridge found\n");
        break;
    case ERR_SELECT_DEVICE:
        printf("selected board not found\n");
        break;
    case ERR_MEM_ALLOC:
        printf("memory allocation error\n");
        break;
    case ERR_MEM_LOCK:
        printf("DMA buffer lock error\n");
        break;
    case ERR_I2CINIT_BUSY:
        printf("I2C bus initialization error\n");
        break;
    case ERR_I2CINIT_ERR:
        printf("I2C bus initialization error\n");
        break;
    case ERR_TIMEOUT:
        printf("internal access timeout\n");
        break;
    case ERR_LOADING_DLL:
        printf("error loading DLL\n");
        break;
    case ERR_FUNCADDR:
        printf("error getting a DLL func\n");
        break;
    case ERR_DLL:
        printf("error accessing DLL\n");
        break;
    case ERR_NO_REGISTERED_DEV:
        printf("unable to register boards\n");
        break;
    case ERR_I2C:
        printf("internal I2C bus error\n");
        break;
    case ERR_I2CTO:
        printf("internal I2C timeout\n");
        break;
    case ERR_EEPROM:
        printf("EEPROM programming error\n");
        break;
    case ERR_XPTO:
        printf("internal crosspoint switch timeout\n");
        break;
    case ERR_ILLEGAL_MODE:
        printf("illegal value in MODE structure\n");
        break;
    case ERR_INTERNAL_OBJ:
        printf("internal obj allocation error\n");
        break;
    case ERR_ACQ_TIMEOUT:
        printf("acquisition timeout\n");
        break;
    case ERR_INTERRUPT:
        printf("interrupt init error\n");
        break;
    case ERR_ILLEGAL_INDEX:
        printf("illegal board index\n");
        break;
    case ERR_PCI:
        printf("PCI access error\n");
        break;
    case ERR_BOARDADDR:
        printf("board address mapping failed\n");
        break;
    case ERR_BOARDREG:
        printf("error registering the board\n");
        break;
    case ERR_ITEMS:
        printf("board has no items\n");
        break;
    case ERR_ALREADYOPEN:
        printf("board already open\n");
        break;
    case ERR_FIFOEMPTY:   
        printf("data FIFO empty\n");
        break;
    case ERR_DATA_ACCESS_TO:
        printf("internal process data access timeout\n");
        break;
    case ERR_DEVICEINFO:
        printf("driver error\n");
        break;
    case ERR_DSP_INIT:
        printf("dsp init\n");
        break;
    case ERR_BOARD_NOT_FOUND:
        printf("board not found\n");
        break;
    case ERR_ILLEGAL_BOARD:
        printf("illegal board index\n");
        break;
    case ERR_ILLEGAL_CHAN:
        printf("illegal channel #\n");
        break;
    case ERR_NO_FREE_BUFFER:
        printf("no free buffers\n");
        break;
    case ERR_DSP_START_TO:
        printf("DSP start timeout\n");
        break;
    case ERR_DSP_CMD_TO:
        printf("HOST-to-DSP command response timeout 40\n");
        break;
    case ERR_DSP_CMD_BUSY:
        printf("HOST-to-DSP command send error (DSP busy)\n");
        break;
    case ERR_DSP_DATA_TO:
        printf("DSP timeout while receiving host data\n");
        break;
    case ERR_DSP_PARSE_DATA:
        printf("DSP parse error\n");
        break;
    case ERR_ALLOC_FIFO:
        printf("error allocating FIFO\n");
        break;
    case ERR_FIFO_EMPTY:
        printf("FIFO contains no data\n");
        break;
    case ERR_DSP_INVALID_ADDR:
        printf("invalid address passed to DSP\n");
        break;
    case ERR_DSP_DATA_TO_CAP:
        printf("data timeout cap\n");
        break;
    case ERR_ALREADY_CAPTURING:
        printf("channel already capturing (StartCapture attempted)\n");
        break;
    case ERR_NOT_CAPTURING:
        printf("channel not capturing (StopCapture attempted)\n");
        break;
    case ERR_ILLEGAL_COMPQUALITY:
        printf("illegal compression quality value 50\n");
        break;
    case ERR_ILLEGAL_FRAMERATE:
        printf("illegal frame rate\n");
        break;
    case ERR_ILLEGAL_TVFORMAT:
        printf("illegal TV format\n");
        break;
    case ERR_ILLEGAL_JPGSIZE:
        printf("illegal jpeg size\n");
        break;
    case ERR_FW_FILE_OPEN:
        printf("error opening firmware file\n");
        break;
    case ERR_FW_FILE_READ:
        printf("error reading firmware file\n");
        break;
    case ERR_ILLEGAL_DSP_INDEX:
        printf("illegal dsp index\n");
        break;
    case ERR_ILLEGAL_BUF:
        printf("illegal buffer index\n");
        break;
    case ERR_ADDR_NOT_CONFIG:
        printf("DSP address not configured on host\n");
        break;
    case ERR_ILLEGAL_CAPMODE:
        printf("illegal capture mode\n");
        break;
    default:
        printf("unspecified error %d\n", ecode);
        break;
    }
    return;
}
// error info
static void getPrintErrorInfo(struct BOARD *b)
{
    int j, k;
    int errors;
    int dsp;
    EXTERR extErr[SYS_EXTERR_BUF_SIZE];
    for (dsp=0; dsp< 4; dsp++) {
        for (j = 0; j < 4; j++) {
            if (S817_GetExtErrInfo(b->board, dsp, j, extErr) != 0) {
                return;
            }
            errors = extErr[0].tick;
            if (errors) {
                printf("DSP %d chan %d Errors %d\n", dsp, j, errors);
                errors = (errors <= SYS_EXTERR_BUF_SIZE - 1) ? errors : SYS_EXTERR_BUF_SIZE - 1;
                for (k = 0; k < errors; k++) {
                    printf ("%08X  %04d  %08X  %08X\n", extErr[k + 1].tick, extErr[k + 1].ecode,
                            extErr[k + 1].reserved[0], extErr[k + 1].reserved[1]);
                }
            }
        }
        // PCI thread errors
        if (S817_GetExtErrInfo(b->board, dsp, 4, extErr) != 0) {
            return;
        }
        errors = extErr[0].tick;
        if (errors) {
            printf("DSP %d chan PCI; Errors %d\n", dsp, errors);
            errors = (errors <= SYS_EXTERR_BUF_SIZE - 1) ? errors : SYS_EXTERR_BUF_SIZE - 1;
            for (k = 0; k < errors; k++) {
                printf ("%08X  %04d  %08X  %08X\n", extErr[k + 1].tick, extErr[k + 1].ecode,
                        extErr[k + 1].reserved[0], extErr[k + 1].reserved[1]);
            }
        } else {
            printf("exterrinfo: no errors found\n");
        }
    }
    getSysStatus(b);
    return;
}


static void getSysStatus(struct BOARD *b)
{
    int jj;
    int dsp;
    SYS_STATUS st;
    for (dsp=0; dsp< 4; dsp++) {
        S817_GetStatusInfo(b->board, dsp, &st);
        printf("sysstat dsp %d\n", dsp);
        for (jj=0;jj<4;jj++) {
            printf("[%d] sysStat lastCmd %d ticks %d reserved: 0x%02x 0x%02x 0x%02x 0x%02x\n",
                   jj, st.chanData[jj].lastCmd, 
                   st.chanData[jj].chanTick,
                   st.reserved[0],
                   st.reserved[1],
                   st.reserved[2],
                   st.reserved[3]);
            printf("[%d] sysStat lockPend %d putToSleep %d fVidStat %d noFreeJPG %d noFreeBmp %d\n",
                   jj, st.chanData[jj].lockPend,
                   st.chanData[jj].putToSleep,
                   st.chanData[jj].fvidStat,
                   st.chanData[jj].noFreeBufJpg,
                   st.chanData[jj].noFreeBufBmp);
        } 
        printf("sys stat fw %u.%u ", (st.fwVer >> 16), (st.fwVer & 0xff));
        printf("tot irq %u\n", st.reserved[0]);
    }
    return;
}

// save JPEG to file
static int save_jpg_file(void *pBuf, int flen, char *fname, int bloud)
{

    FILE *fin;
    int ecode = 0;
    int rc;
    char fpath[260];
    if (strstr(fname, ".jpg") == NULL)
        snprintf(fpath, 260, "%s.jpg", fname);
    else
        strncpy(fpath, fname, 260);
        
    if (bloud)
        printf("savefile %s, len:%d\n", (char *)fpath, flen);

    fin = fopen(fpath, "wb+");

    if (fin != NULL) {
        rc = fwrite( pBuf, 1, flen, fin);
        if (rc <= 0)
            ecode = rc;
        fflush(fin);
        fclose(fin);
    } else {
        if (bloud)
            printf("error\n");
        ecode = -1;
    }
    return ecode;
}

static int save_image_ppm(const unsigned char *image, const char *szFilename, int width, int height)
{
    char ppmheader[256];
    FILE *fptr;
    int rlen;
    int clen;
    int i;
    int rc;
    char fpath[260];
    rlen = width;
    clen = height;
    
    if (strstr(szFilename, ".ppm") == NULL)
        snprintf(fpath, 260, "%s.ppm", szFilename);
    else
        strncpy(fpath, szFilename, 260);

    fptr = fopen(fpath, "wb");
    if (fptr == NULL) {
        printf("could not open file\n");
        return -1;
    }
    sprintf( ppmheader, "P5\n#ppm image\n%d %d\n255\n", width, height);
    rc = fwrite( ppmheader, 1, strlen(ppmheader), fptr);
    if (rc <= 0) {
        fclose(fptr);
        return rc;
    }
            
    // write out the rows
    for (i=0; i < clen; i++) {
        rc = fwrite( &image[rlen*i], 1, rlen, fptr);
        if (rc <= 0) {
            fclose(fptr);
            return rc;
        }
    }
    printf("saved in PPM format to filename: %s\n", fpath);
    fclose(fptr);
    return 0;
}

